/* 
    ssft - is a simple file transfer program that use SSL (Secure 
    Socket Layer) encryption to secure the transfers - and 
    signed certificates to authenticate users and servers. 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    Copyright 2001 by Jarle (jgaa) Aase, jgaa@jgaa.com
    http://www.jgaa.com

*/


#include "StdAfx.h"
#include "ssft.h"
#include <assert.h>

using namespace std;

typedef list<string> arglist_t;

int do_client_xmit(int argc, char **argv)
{
    int index = 0;
    string dst_path;
    arglist_t src_list;

    if (argc < 2)
    {
        log << "No files specifield" << ENDL;
        return -1;
    }

    // Validate src files

    bool client_to_server = true;

    for(index = 0; index < (argc -1); index++)
    {
        bool my_direction = (strchr(argv[index], ':') == NULL);

        if (0 == index)
            client_to_server = my_direction;
        else if (my_direction != client_to_server)
        {
            log << LOG_DATE 
                << "Error parsing arguments: Source files must "
                "be either on local macine or on the server. They "
                "can not be mixed."
                << ENDL;
            return -1;
        }

        if (client_to_server)
        {
            src_list.push_back(argv[index]);
        }
        else
        {
            // Get hostname
            const char *p = argv[index];
            string my_hostname;

            while(*p && (':' != *p))
                my_hostname += *p++;

            if (my_hostname.empty())
            {
                log << LOG_DATE 
                    << "Error parsing arguments: Missing hostname! "
                    << argv[index]
                    << ENDL;
                return -1;
            }

            if (hostname.empty())
            {
                hostname = my_hostname;
            }
            else if (hostname != my_hostname)
            {
                log << LOG_DATE 
                    << "Error parsing arguments: Can only use one server in one "
                    "operation. You have already specified the server: "
                    << hostname
                    << ". "
                    << argv[index]
                    << ENDL;
                return -1;
            }

            if (!*p || !*++p)
            {
                log << LOG_DATE 
                    << "Error parsing arguments: No path after hostname! "
                    << argv[index]
                    << ENDL;
                return -1;
            }
            
            src_list.push_back(p);
        }
    }

    // Validate dst path
    {
        bool my_direction = (strchr(argv[argc -1], ':') == NULL);

        if (my_direction == client_to_server)
        {
            if (client_to_server)
            {
                log << LOG_DATE 
                    << "Why would you use ssft to copy files on "
                    "your local system? Use the "
#ifdef  WIN32
                    "copy "
#else
                    "cp "
#endif
                    "command for that!"
                    << ENDL;
            }
            else
            {
                 log << LOG_DATE 
                    << "Server to server (proxy) transfers are not possible."
                    << ENDL;
            }
            return -1;
        }

        const char *p = argv[argc -1];

        if (client_to_server)
        {
            while(*p && (*p != ':'))
                hostname += *p++;
            
            if (!*p || !*++p)
            {
                log << LOG_DATE 
                    << "Error parsing arguments: No path after hostname! "
                    << argv[index]
                    << ENDL;
                return -1;
            }
        }

        dst_path = p;

        if (src_list.empty())
        {
            log << LOG_DATE << "Error: No source file(s) specified" << ENDL;
            return -1;
        }

        if (hostname.empty())
        {
            log << LOG_DATE << "Error: No host is specified" << ENDL;
            return -1;
        }

        if (dst_path.empty())
        {
            log << LOG_DATE << "Error: No target path is specified" << ENDL;
            return -1;
        }

        
        if (!resolve_hostname())
            return -1;
    }

    SSL_load_error_strings();
    SSLeay_add_ssl_algorithms();
    SSL_METHOD *meth = SSLv23_client_method();
    SSL_CTX *ctx = SSL_CTX_new(meth);
    if (!ctx) 
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_CTX_new() failed " 
            << ssl_err << ENDL;
        
        SSL_CTX_free (ctx);
        return -2;
    }

    if (SSL_CTX_load_verify_locations(ctx, rootcert_file.c_str(), NULL) <= 0)
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_CTX_load_verify_locations(\""
	    << rootcert_file
	    << "\") failed " 
            << ssl_err << ENDL;
        
        SSL_CTX_free (ctx);
        return -3;
    }

    if (SSL_CTX_set_default_verify_paths(ctx) <= 0)
    {
        SslError ssl_err;
        
        log << LOG_DATE  
            << "ERROR - SSL_CTX_load_verify_locations() failed " 
            << ssl_err << ENDL;        

        SSL_CTX_free (ctx);
        return -4; 
    }

    if (SSL_CTX_use_certificate_file(ctx, cli_cert_file.c_str(), SSL_FILETYPE_PEM) <= 0) 
    {
        SslError ssl_err;

        log << LOG_DATE  
            << "ERROR - SSL_CTX_use_certificate_file() failed " 
            << ssl_err << ENDL;
        
        SSL_CTX_free (ctx);
        return -5;
    }
    
    if (SSL_CTX_use_PrivateKey_file(ctx, cli_key_file.c_str(), SSL_FILETYPE_PEM) <= 0) 
    {
        SslError ssl_err;

        log << LOG_DATE  
            << "ERROR - SSL_CTX_use_certificate_file() failed " 
            << ssl_err << ENDL;
        
        return -6;
    }

    if (debug_flag)
        log << "Using client certificate: " << cli_cert_file << ENDL;

    if (SSL_CTX_check_private_key(ctx) <= 0) 
    {
        SslError ssl_err;

        log << LOG_DATE  
            << "ERROR - Private key does not match the certificate public key " 
            << ENDL;
        
        SSL_CTX_free (ctx);
        return -7;
    }
    
    SOCKET server = create_socket(false);
    
    log << "Connecting to "
        << inet_ntoa(pin->sin_addr) 
        << ":" 
        << ntohs(pin->sin_port) 
        << ENDL;
    
    if (connect(server, &socaddr, sizeof(socaddr)) == -1)
    {
        log << LOG_DATE  
            << "failed to connect to server " 
            << GET_SOCKET_ERROR
            << ENDL;

        SSL_CTX_free (ctx);
        return -8;
    }

    ClientConnection my_client(server, ctx);
    if (!my_client.Create())
    {
        SSL_CTX_free(ctx);
        return -9;
    }

    log << LOG_DATE  << "connected to server" << ENDL;
    
    for(arglist_t::const_iterator P = src_list.begin()
        ; P != src_list.end()
        ; ++P)
    {
        try
        {
            if (client_to_server)
                my_client.SendFile(P->c_str(), dst_path);
            else
                my_client.RecvFile(P->c_str(), dst_path);
        }
        catch(...)
        {
            log << LOG_DATE  
                << "Transfer failed" << ENDL;

            SSL_CTX_free(ctx);
            return -10;
        }
    }

    try
    {
        my_client.Quit();
    }
    catch(...)
    {
    }

    my_client.Close();
    
    SSL_CTX_free (ctx);

    log << "Successfully done." << ENDL;
    
    return 0;
}


